import pandas as pd 
from scipy.io import loadmat
import numpy as np
import os

class LoadData():
    """class responsible for loading various data formats"""

    def __init__(self, path, path_file):
        """set the name and location of the file"""

        # directory where the file is located
        self.path = path
        # data file name
        self.path_file = path_file

        return None

    def load_csv(self):
        """function that opens csv files"""

        csv_path = os.path.join(self.path, self.path_file)

        return pd.read_csv(csv_path)
    
    def load_excel(self):
        """function that opens excel files"""

        excel_path = os.path.join(self.path, self.path_file)

        return pd.read_excel(excel_path)

    def load_json(self):
        """function that opens json files"""

        json_path = os.path.join(self.path, self.path_file)

        return pd.read_json(json_path)
    
    def load_mat(self):
        """function that opens matlab files"""

        mat_path = os.path.join(self.path, self.path_file)

        return loadmat(mat_path)
    
    def load_txt(self):
        """function that opens text files"""

        txt_path = os.path.join(self.path, self.path_file)

        return pd.read_csv(txt_path, header = None)

    def convert_numpy_array(self, data):
        """turns a data file into a numpy array"""
        
        data = np.array(data)

        return data


import sys
sys.path.append('../MachineLearningPackage/')

from pyLearnAlgorithms.load_data import LoadData
from pyLearnAlgorithms.split_data import SplitData

json_file = LoadData('tests/datasets', 'JSONfile.json')

jsonfile = json_file.load_json()

data = SplitData(jsonfile)

train, test, val = data.split_train_test_val(0.2, 0.3)

print('data size: ', jsonfile.size)
print('train size: ',train.size)
print('test size: ', test.size)
print('val size: ', val.size)

print(type(train))

train = json_file.convert_numpy_array(train)
test = json_file.convert_numpy_array(test)
val = json_file.convert_numpy_array(val)

print(type(train))



import sys
sys.path.append('../MachineLearningPackage/')

from pyLearnAlgorithms.load_data import LoadData

json_file = LoadData('tests/datasets', 'JSONfile.json')

jsonfile = json_file.load_json()

print(jsonfile)

jsonfile = json_file.convert_numpy_array(jsonfile)

print('\n', jsonfile)



import numpy as np
from scipy import optimize

class LinearRegression():
    """class for building linear regression models"""

    def __init__(self, X, y):
        """stores predictive attributes and labels"""

        # predictive attributes
        self.X = X
        # values to be predicted
        self.y = y

        return None
    
    def linearRegCostFunction(self, theta = [1,1], lambda_= 0.0):
        """cost function for a varied hypothesis (with regularization)"""

        theta = np.array(theta)
        m = self.y.size 
        X_bias = np.concatenate([np.ones((m, 1)), self.X], axis=1)
        J = 0
        grad = np.zeros(theta.shape)
        h = X_bias.dot(theta)
        J = (1/(2*m))*np.sum(np.square(h-self.y))+(lambda_/(2*m))*np.sum(np.square(theta[1:]))
        grad = (1/m)*(h-self.y).dot(X_bias)
        grad[1:] = grad[1:] + (lambda_/m)*theta[1:]

        return J, grad

    def trainLinearReg(self, lambda_ = 0.0, maxiter = 200):
        """training the model for minimizing the cost function and modifying theta parameters"""

        model = LinearRegression(self.X, self.y)
        m = self.y.size 
        X_bias = np.concatenate([np.ones((m, 1)), self.X], axis=1)
        teta_inicial = np.zeros(len(X_bias[0, :]))
        costFunction = lambda t: model.linearRegCostFunction(t, lambda_)
        options = {'maxiter': maxiter}
        res=optimize.minimize(costFunction,teta_inicial,jac=True,method='TNC',options=options)

        return res

    def learningCurve(self, Xval, yval, lambda_= 0.0):
        """computes training error and validation error for different data extracts"""

        model = LinearRegression(self.X, self.y)
        m = self.y.size
        erro_treinamento = np.zeros(m)
        erro_validacao   = np.zeros(m)
        for i in range(1, m + 1):
            model = LinearRegression(self.X[:i], self.y[:i])
            teta_t = model.trainLinearReg()
            erro_treinamento[i-1],_=model.linearRegCostFunction(teta_t.x)
            model = LinearRegression(Xval, yval)
            erro_validacao[i-1],_=model.linearRegCostFunction(teta_t.x)
        
        return erro_treinamento, erro_validacao
    
    def predict(self, Xtest):
        """makes the prediction for the test values"""

        model = LinearRegression(self.X, self.y)
        optimized_values = model.trainLinearReg(lambda_= 0.0)
        X_bias = np.concatenate([np.ones((Xtest.size, 1)), Xtest], axis=1)
        pred = np.dot(X_bias, optimized_values.x)

        return pred
    
    def predict_normal_equation(self, Xtest):
        """makes the prediction for the test values with normal equation"""

        model = LinearRegression(self.X, self.y)
        _, grad = model.normal_equation()
        X_bias = np.concatenate([np.ones((Xtest.size, 1)), Xtest], axis=1)
        pred = np.dot(X_bias, grad)

        return pred
    
    def normal_equation(self):
        """computes the linear regression model using the normal equation (least quadratic)"""

        m = self.y.size 
        X_bias = np.concatenate([np.ones((m, 1)), self.X], axis=1)
        grad = np.linalg.inv((X_bias.T.dot(X_bias))).dot(X_bias.T.dot(self.y))
        model = LinearRegression(self.X, self.y)
        J, _ = model.linearRegCostFunction(grad, lambda_ = 0.0)
    
        return J, grad

    



import sys
sys.path.append('../MachineLearningPackage/')

from pyLearnAlgorithms.load_data import LoadData

excel_file = LoadData('tests/datasets', 'Adresses.xlsx')

adresses = excel_file.load_excel()

print(adresses)

adresses = excel_file.convert_numpy_array(adresses)

print('\n', adresses)


from matplotlib import pyplot as plt
import numpy as np
from pyLearnAlgorithms.linear_regression import LinearRegression
from pyLearnAlgorithms.polynomial_regression import PolynomialRegression

class GraphView():
    """graphic display class of machine learning models"""

    def __init__(self, X, y):
        """stores predictive attributes and labels"""

        # predictive attributes
        self.X = X
        # values to be predicted
        self.y = y
    
        return None

    def view_data(self, xlabel, ylabel, title):
        """data visualization"""

        plt.figure(figsize = (10, 5))
        plt.plot(self.X, self.y, 'ro', ms=10, mec='k', mew=1) 
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(title)

        return None

    def model_linear(self, xlabel, ylabel, title):
        """plots the graph of the linear regression model"""

        model = LinearRegression(self.X, self.y)
        values = model.trainLinearReg()
        m = self.y.size 
        plt.figure(figsize = (10,5))
        X_bias = np.concatenate([np.ones((m, 1)), self.X], axis=1)
        plt.plot(self.X, self.y, 'ro', ms=10, mec='k', mew=1.5)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(title)
        plt.plot(self.X, np.dot(X_bias, values.x), '--', lw=2)

        return None

    def model_linear_normal_equation(self, xlabel, ylabel, title):
        """plots the graph of the linear regression model with normal equation"""

        model = LinearRegression(self.X, self.y)
        _, grad = model.normal_equation()
        m = self.y.size 
        plt.figure(figsize = (10,5))
        X_bias = np.concatenate([np.ones((m, 1)), self.X], axis=1)
        plt.plot(self.X, self.y, 'ro', ms=10, mec='k', mew=1.5)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(title)
        plt.plot(self.X, np.dot(X_bias, grad), '--', lw=2)

        return None
    
    def model_poly(self, xlabel, ylabel, title, polynomial_degree, lambda_ = 0.0, maxiter = 200):
        """plots the graph of the polynomial regression model"""

        x = np.arange(np.min(self.X) - 15, np.max(self.X) + 25, 0.05).reshape(-1, 1)
        model_poly = PolynomialRegression(self.X, self.y)
        X_poly = model_poly.polyFeatures(x, polynomial_degree)
        data_poly = model_poly.polyFeatures(self.X, polynomial_degree)
        _, mu, sigma = model_poly.featureNormalize(data_poly)
        X_poly -= mu
        X_poly /= sigma
        X_poly = np.concatenate([np.ones((x.shape[0], 1)), X_poly], axis=1)
        optimized_values = model_poly.trainLinearReg(polynomial_degree, lambda_, maxiter)
        plt.figure(figsize = (10, 5))
        plt.plot(self.X, self.y, 'ro', ms=10, mew=1.5, mec='k')
        plt.plot(x, np.dot(X_poly, optimized_values.x), '--', lw=2)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(title)
        plt.ylim([np.min(self.y) - 15, np.max(self.y) + 25])
        
        return None
    
    def validation_curve_poly(self, X_poly, y,Xval, yval, lambda_vec):
        """graphically analyzes the best value of the lambda hyperparameter for normalization"""

        model = PolynomialRegression(self.X, self.y)
        lambda_vec, error_train, error_val = model.validationCurve(X_poly,y,Xval,yval,lambda_vec)
        plt.figure(figsize = (10, 5))
        plt.plot(lambda_vec, error_train, '-o', lambda_vec, error_val, '-o', lw = 2)
        plt.legend(['Train', 'Cross Validation'])
        plt.xlabel('lambda')
        plt.ylabel('Error')

        return None
    
    def learning_curve(self, error_train, erro_val):
        """learning curve for bias vs variance analysis"""

        m = self.y.size 
        plt.figure(figsize = (10, 5))
        plt.plot(np.arange(1, m+1), error_train, np.arange(1, m+1), erro_val, lw=2)
        plt.title('Learning Curve')
        plt.legend(['Training', 'Cross Validation'])
        plt.xlabel('Number of trainable examples')
        plt.ylabel('Error')

        return None
    
    def predicted_values(self, Xtest, ytest, pred):
        """visualization of expected data with that obtained"""

        plt.figure(figsize = (10,5))
        plt.scatter(Xtest, ytest, label = 'correct values')
        plt.scatter( Xtest, pred, label = 'predict values')

        plt.legend()
        plt.grid(True)
        plt.box(False)
        plt.title('Predicted Values')

        return None

    
    





import sys
sys.path.append('../pyLearnAlgorithms/')

from pyLearnAlgorithms.load_data import LoadData
from pyLearnAlgorithms.split_data import SplitData

matlab_file = LoadData('tests/datasets', 'WaterLevel.mat')

waterlevel = matlab_file.load_mat()

data = SplitData(waterlevel)

train, test = data.split_train_test(0.2)



import numpy as np
import pandas as pd

class SplitData():
    """class responsible for dividing the data"""
    
    def __init__(self, data):
        # data to be divided
        self.data = data
        
        return None
    
    def split_train_test(self, test_ratio):
        """divides data for training and testing"""

        if type(self.data) == dict:
            # matlab file is not yet supported
            print('unsupported data format')

            return None, None

        shuffled_indices = np.random.permutation(len(self.data))
        test_set_size = int(len(self.data) * test_ratio)
        test_indices = shuffled_indices[:test_set_size]
        train_indices = shuffled_indices[test_set_size:]

        return self.data.iloc[train_indices], self.data.iloc[test_indices]
    
    def split_train_test_val(self, test_ratio, val_ratio):
        """divides data for training, validation and testing"""

        if type(self.data) == dict:
            # matlab file is not yet supported
            print('unsupported data format')

            return None, None, None

        shuffled_indices = np.random.permutation(len(self.data))
        test_set_size = int(len(self.data) * test_ratio)
        val_set_size = int(len(self.data) * val_ratio)
        test_indices = shuffled_indices[:test_set_size]
        val_indices = shuffled_indices[test_set_size:(val_set_size + test_set_size)]
        train_indices = shuffled_indices[(val_set_size + test_set_size):]

        return self.data.iloc[train_indices], self.data.iloc[test_indices], self.data.iloc[val_indices]




import sys
sys.path.append('../MachineLearningPackage/')

from pyLearnAlgorithms.load_data import LoadData

matlab_file = LoadData('tests/datasets', 'WaterLevel.mat')

waterlevel = matlab_file.load_mat()

print(waterlevel)

waterlevel = matlab_file.convert_numpy_array(waterlevel)

print('\n', waterlevel)



from setuptools import setup

with open("README.md", "r") as fh:
    readme = fh.read()

setup(name='pyLearnAlgorithms',
    version = '0.2.0',
    url='https://github.com/Alyssonmach/pyLearnAlgorithms',
    license='MIT License',
    author='Alysson Machado',
    long_description=readme,
    long_description_content_type="text/markdown",
    author_email='alysson.barbosa@ee.ufcg.edu.br',
    keywords=['Machine Learning', 'Regression', 'Classification'],
    description=u'Python Machine Learning algorithms based on Numpy, Scipy and Panda.',
    packages=['pyLearnAlgorithms'],
    install_requires=['numpy', 'scipy', 'pandas', 'matplotlib'],)


import numpy as np
from pyLearnAlgorithms.linear_regression import LinearRegression 
from scipy import optimize

class PolynomialRegression():
    """class for building polynomial regression models"""
    
    def __init__(self, X, y):
        """stores predictive attributes and labels"""

        # predictive attributes
        self.X = X
        # values to be predicted
        self.y = y

        return None

    def polyFeatures(self, X, polynomial_degree = 1):
        """performs polynomial mapping on predictive attributes"""

        X_poly = np.zeros((X.shape[0], polynomial_degree))
        for i in range(polynomial_degree):
            X_poly[:, i] = X[:, 0] ** (i + 1)

        return X_poly
    
    def featureNormalize(self, X):
        """performs normalization on predictive attributes by normal distribution"""
    
        mu = np.mean(X, axis = 0)
        X_norm = (X - mu)
        sigma = np.std(X_norm, axis = 0, ddof = 1)
        X_norm = X_norm / sigma
        m = self.y.size 
        X_norm = np.concatenate([np.ones((m, 1)), X_norm], axis=1)
        
        return X_norm, mu, sigma

    def prepareExtracts(self, extractX, extractY, polynomial_degree):
        """performs normalization and polynomial mapping on other data extracts"""

        model = PolynomialRegression(self.X, self.y)
        X_poly = model.polyFeatures(extractX, polynomial_degree)
        data_poly = model.polyFeatures(self.X, polynomial_degree)
        _, mu, sigma = model.featureNormalize(data_poly)
        X_poly -= mu
        X_poly /= sigma
        X_poly = np.concatenate([np.ones((extractY.size, 1)), X_poly], axis=1)

        return X_poly

    def trainLinearReg(self, polynomial_degree, lambda_ = 0.0, maxiter = 200):
        """training the model for minimizing the cost function and modifying theta parameters"""

        model_poly = PolynomialRegression(self.X, self.y)
        X_poly = model_poly.polyFeatures(self.X, polynomial_degree)
        X_poly, _, _ = model_poly.featureNormalize(X_poly)
        teta_inicial = np.zeros(X_poly[0,:].size)
        X_poly = np.delete(X_poly, 0, 1)
        model_linear = LinearRegression(X_poly, self.y)
        costFunction = lambda t: model_linear.linearRegCostFunction(t, lambda_)
        options = {'maxiter': maxiter}
        res=optimize.minimize(costFunction,teta_inicial,jac=True,method='TNC',options=options)

        return res
    
    def learningCurve(self, X_poly, y, Xval, yval, lambda_ = 0.0):
        """computes training error and validation error for different data extracts"""
        
        m = y.size
        error_train = np.zeros(m)
        error_val   = np.zeros(m)
        for i in range(1, m + 1):
            teta_t = self.__trainLinearReg(X_poly[:i], y[:i], lambda_)
            error_train[i-1],_=self.__linearRegCostFunction(X_poly[:i], y[:i],teta_t.x, lambda_)
            error_val[i - 1], _ = self.__linearRegCostFunction(Xval, yval, teta_t.x, lambda_)
            
        return error_train, error_val
    
    def validationCurve(self,X_poly,y,Xval,yval,lambda_vec=[0,0.001,0.003,0.01,0.03,0.1,0.3,1,3,10]):
        """analyzes the lambda values for the best normalization"""

        error_train = np.zeros(len(lambda_vec))
        error_val = np.zeros(len(lambda_vec))
        for i in range(len(lambda_vec)):
            lambda_try = lambda_vec[i]
            teta_t = self.__trainLinearReg(X_poly, y, lambda_ = lambda_try)
            error_train[i], _ = self.__linearRegCostFunction(X_poly, y, teta_t.x, lambda_ = 0)
            error_val[i], _ = self.__linearRegCostFunction(Xval, yval, teta_t.x, lambda_ = 0)

        return lambda_vec, error_train, error_val
    
    def predict(self, Xtest, X_poly, y, lambda_ = 0.0):
        """makes the prediction for the test values"""

        teta_t = self.__trainLinearReg(X_poly, y, lambda_)
        pred = np.dot(Xtest, teta_t.x)

        return pred 
    
    def __linearRegCostFunction(self, X, y, theta = [1,1], lambda_= 0.0):
        """cost function for a varied hypothesis (with regularization)"""

        theta = np.array(theta)
        m = y.size 
        J = 0
        grad = np.zeros(theta.shape)
        h = X.dot(theta)
        J = (1/(2*m))*np.sum(np.square(h-y))+(lambda_/(2*m))*np.sum(np.square(theta[1:]))
        grad = (1/m)*(h-y).dot(X)
        grad[1:] = grad[1:] + (lambda_/m)*theta[1:]

        return J, grad
    
    def __trainLinearReg(self,X, y, lambda_ = 0.0, maxiter = 200):
        """training the model for minimizing the cost function and modifying theta parameters"""

        theta = np.zeros(X.shape[1])
        costFunction = lambda t: self.__linearRegCostFunction(X, y, t, lambda_)
        options = {'maxiter': maxiter}
        res = optimize.minimize(costFunction,theta,jac=True,method='TNC',options=options)

        return res

    



    


import sys
sys.path.append('../MachineLearningPackage/')

from pyLearnAlgorithms.load_data import LoadData

csv_file = LoadData('tests/datasets', 'CreditRiscs.csv')

creditriscs = csv_file.load_csv()

print(creditriscs)

creditriscs = csv_file.convert_numpy_array(creditriscs)

print('\n', creditriscs)
